module genumSampleDataStructure;

//--------------------------------------------------------------
// データ構造の例をいくつか
//--------------------------------------------------------------

class List(T)
{
   private T        car;
   private List!(T) cdr;
}

class BinTree(T)
{
   private T           value;
   private BinTree!(T) left;
   private BinTree!(T) right;
}

class Tree(T)
{
  static class Node
  {
     private T      value;
     private Node[] children;

     void add_child(Node[] ns...) { children ~= ns; }
  }
  private Node root;

  Node create_node(T v) 
  {
     Node n = new Node;
     n.value = v;
     return n;
  }
  void set_root(Node r) { root = r; }
}


class Set(T) if( is(typeof(valueof!(T) < valueof!(T))) )
{
   // 実は single-rotation までしか実装してないので
   // 全然AVL木じゃない。時間がなかった。今は反省している。
   static class AVLTree
   {
      private int     height;

      private AVLTree left;
      private T       value;
      private AVLTree right;

      mixin SimpleConstructor;

      private void recalcHeight() {
         int hl = left ? left.height : 0;
         int hr = right ? right.height : 0;
         height = (hl>hr ? hl : hr) + 1;
      }
   }

   AVLTree root = null;

   void add( T t )
   {
      root = insert(t, root);
   }

   private AVLTree insert( T t, AVLTree node )
   {
      if( node is null )
         return new AVLTree(1, null, t, null);

      if( t < node.value )
      {
         node.left = insert(t, node.left);
         if( node.left.height
             >= (node.right ? node.right.height : 0)+2 )
         {
             auto newRoot = node.left;
             node.left = newRoot.right;
             newRoot.right = node;
             node.recalcHeight();
             newRoot.recalcHeight();
             return newRoot;
         }
      }

      if( node.value < t )
      {
         node.right = insert(t, node.right);
         if( node.right.height
             >= (node.left ? node.left.height : 0)+2 )
         {
             auto newRoot = node.right;
             node.right = newRoot.left;
             newRoot.left = node;
             node.recalcHeight();
             newRoot.recalcHeight();
             return newRoot;
         }
      }

      return node;
   }

   bool has( T t )
   {
       return has(t, root);
   }

   private bool has( T t, AVLTree node )
   {
       return (node !is null)
   && (node.value==t || has(t, node.left) || has(t, node.left));
   }
}

//--------------------------------------------------------------
// ちょっとしたユーティリティ
//--------------------------------------------------------------

// こんすとらくためーかー

template SimpleConstructor()
{
   this( typeof(this.tupleof) xs )
     { this.tupleof = xs; }
}

T Constructor_impl(T,P...)(P p)
{
   T t = new T;
   t.tupleof[0..p.length] = p;
   return t;
}

template Constructor(T, int N = T.tupleof.length)
{
   alias Constructor_impl!(
      T, typeof(T.tupleof)[0..N]) Constructor;
}

// 型からその型をもつ式を作る
//  - valueof!(int) は int 型の式
//
// なぜ直接 T.init と書かないかというと、
// T が class のときに
//   is(typeof(T.init < T.init))
// は false になる！！！ので
// ※ T.init は null で、この式はnull.opCmp(null)<0 になって
//    nullのメソッド呼び出しはコンパイル時にはねられる…
// これをごまかすため

T valueof(T)() { return T.init; }
